#include "block.h"
#include <basedef.h>
#include <vmm.h>
#include <debug.h>
#include <vxdwraps.h>
#include <vwin32.h>
#include <memory.h>
#pragma intrinsic( memcpy )
#pragma intrinsic( memset )

#define MAX_PROCESSES   16

#ifdef SEMAPHORE
#define BLOCK_CREATE       Create_Semaphore( 0 )
#define BLOCK(id)          Wait_Semaphore( id, \
                               BLOCK_THREAD_IDLE | BLOCK_SVC_INTS ) 
#define BLOCK_DESTROY(id)  Destroy_Semaphore( id )
#define SIGNAL(id)         Signal_Semaphore( id )
#else
#define BLOCK_CREATE       (DWORD)HeapAllocate( sizeof(DWORD),  0 )
#define BLOCK(id)          BlockOnID( id, 0 )
#define SIGNAL(id)         SignalID( id )
#define BLOCK_DESTROY(id)  HeapFree( (const void *)id, 0 )

#endif

typedef struct
{
  HANDLE      pid;
  DWORD       BlockOnThis;
  DWORD       Status;
  HTIMEOUT   hndTimeout;
} PROCESS_INSTANCE;

PROCESS_INSTANCE *aProcesses[MAX_PROCESSES];
char dbgbuf[100];

BOOL OnSysDynamicDeviceInit( void );
BOOL OnSysDynamicDeviceExit( void );
DWORD OnW32Deviceiocontrol( PDIOCPARAMETERS p );

PROCESS_INSTANCE  * AllocProcessInstance( HANDLE pid );
PROCESS_INSTANCE  * FindProcessInstance( HANDLE pid );
void FreeProcessInstance( PROCESS_INSTANCE *p );
VOID __stdcall TimeoutHandler(HVM hVM, PCRS pcrs, PVOID RefData, 
                              DWORD Extra);

// functions in ASM modu
extern void TimeoutThunk(void );

BOOL OnSysDynamicDeviceInit()
{
   return TRUE;
}

BOOL OnSysDynamicDeviceExit()
{
   return TRUE;
}

DWORD OnW32Deviceiocontrol(PDIOCPARAMETERS p)
{
   DWORD rc;
   PROCESS_INSTANCE *process;

   switch( p->dwIoControlCode)
   {
   case DIOC_OPEN:
      // get per-process instance struct 

      process = AllocProcessInstance( (HANDLE)p->tagProcess );
      if (!process)
         return (DWORD)0xffffffff;
      rc = 0;  
      break;

   case DIOC_CLOSEHANDLE:
      process = FindProcessInstance( (HANDLE)p->tagProcess );
      FreeProcessInstance( process );
      rc = 0;
      break;

   case DIOC_BLOCK:     
      process = FindProcessInstance( (HANDLE)p->tagProcess );
      if (!process)
         return (DWORD)0xffffffff;

      process->BlockOnThis = BLOCK_CREATE;                    
      if (!process->BlockOnThis)
         return (DWORD)0xffffffff;

      // schedule a timeout to signal us
      process->hndTimeout = Set_Global_Time_Out( TimeoutThunk,   
                           p->lpvInBuffer, (ULONG)process );

      // block here until signalled by scheduled timeout event
      BLOCK( process->BlockOnThis );

      // we've been signalled...
      BLOCK_DESTROY( process->BlockOnThis );

      memcpy( (void *)p->lpvOutBuffer, 
              &process->Status, 
              sizeof(process->Status));
      p->lpcbBytesReturned = sizeof(process->Status);
      rc = 0;

      break;
   
   default:
      rc = 0xFFFFFFFF;
   }
   return rc;
}

PROCESS_INSTANCE  * AllocProcessInstance( HANDLE pid )
{
   int   i;
   PROCESS_INSTANCE *p;

   p = (PROCESS_INSTANCE *)_HeapAllocate( sizeof(PROCESS_INSTANCE), 
                                          0 );
   if (!p)
      return NULL;
   memset( p, 0, sizeof(PROCESS_INSTANCE) );
   p->pid = pid;
   for (i=0; i < MAX_PROCESSES; i++)
   {
      if (aProcesses[i] == NULL)
      {
         aProcesses[i] = p;
         break;
      }
   }
   if (i >= MAX_PROCESSES)
      return NULL;
   else
      return p;
}

void FreeProcessInstance( PROCESS_INSTANCE *p )
{
   int   i;

   for (i=0; i < MAX_PROCESSES; i++)
   {
      if (aProcesses[i]->pid == p->pid)
      {
         _HeapFree( aProcesses[i], 0 );
         aProcesses[i] = NULL;
         break;
      }
   }
}

PROCESS_INSTANCE  * FindProcessInstance( HANDLE pid )
{
   int   i;

   for (i=0; i < MAX_PROCESSES; i++)
   {
      if (aProcesses[i]->pid == pid)
      {
      break;
      }
   }
   if (i >= MAX_PROCESSES)
      return NULL;
   else 
      return aProcesses[i];
}

VOID __stdcall TimeoutHandler(HVM hVM, PCRS pcrs, PVOID RefData, DWORD Extra)
{
   PROCESS_INSTANCE  *process = (PROCESS_INSTANCE *)RefData;

   process->Status = Extra;
   SIGNAL( process->BlockOnThis );
}
