//**************************************************************************
//
//	File Name:  DlgMemory.h
//
//	Class Name(s):  CDlgMemory
//
//	Author:  Theron S. Welch
//
//	Purpose:	These classes work together to provide a way for 
//				dialogs to memorize their options from one session
//				to the next, and (eventually) even persistenly through the 
//				registry.
//
//**************************************************************************

#pragma once

/////////////////////////////////////////////////////////////////////////////

// Abstract base class for windows control types
class CControlMemory
{
	public:
		CControlMemory( int nID ) : m_nID( nID ){}
		virtual ~CControlMemory(){}

		virtual bool RestoreState( CWnd* pParent )=0;
		virtual bool SaveState( CWnd* pParent )=0;

	protected:

		int m_nID;

	private:

		CControlMemory();
};

/////////////////////////////////////////////////////////////////////////////

class CRadioMemory : public CControlMemory
{
	public:
		CRadioMemory( int nID ) : CControlMemory( nID ){ m_nState = -1; }
		bool RestoreState( CWnd* pParent );
		bool SaveState( CWnd* pParent );

	private:
		int m_nState;
};

/////////////////////////////////////////////////////////////////////////////

class CCheckBoxMemory : public CControlMemory
{
	public:
		CCheckBoxMemory( int nID ) : CControlMemory( nID ){ m_nState = -1; }
		bool RestoreState( CWnd* pParent );
		bool SaveState( CWnd* pParent );

	private:
		int m_nState;
};

/////////////////////////////////////////////////////////////////////////////

class CEditBoxMemory : public CControlMemory
{
	public:
		CEditBoxMemory( int nID ) : CControlMemory( nID ){}
		bool RestoreState( CWnd* pParent );
		bool SaveState( CWnd* pParent );

	private:
		CString m_sState;
};

/////////////////////////////////////////////////////////////////////////////

class CSliderMemory : public CControlMemory
{
	public:
		CSliderMemory( int nID ) : CControlMemory( nID ){	m_nPos = 0;	}
		
		bool RestoreState( CWnd* pParent );
		bool SaveState( CWnd* pParent );

	private:
		int m_nPos;
};

/////////////////////////////////////////////////////////////////////////////

class CComboBoxMemory : public CControlMemory
{
	public:
		CComboBoxMemory( int nID ) : CControlMemory( nID ){}
		bool RestoreState( CWnd* pParent );
		bool SaveState( CWnd* pParent );

	private:
		int m_nSelection;
};

/////////////////////////////////////////////////////////////////////////////

class CListBoxMemory : public CControlMemory
{
	public:
		CListBoxMemory( int nID ) : CControlMemory( nID )
		{
			m_nCount = 0;
			m_pSelection = NULL;
		}
		~CListBoxMemory()
		{
			if ( m_pSelection )
				delete [] m_pSelection;
		}

		bool RestoreState( CWnd* pParent );
		bool SaveState( CWnd* pParent );

	private:
		int* m_pSelection;
		int m_nCount;
};

/////////////////////////////////////////////////////////////////////////////

class CDlgMemory
{
	public:

		CDlgMemory();
		virtual ~CDlgMemory();

		void Attach( CWnd* pDlg );

		bool SaveRadioButtons( int nRadioID );
		bool SaveCheckBox( int nCheckBoxID );
		bool SaveEditBoxText( int nEditBoxID );
		bool SaveSliderInfo( int nSliderID );
		bool SaveComboBoxSelection( int nComboBoxID );
		bool SaveListBoxSelection( int nListBoxID );

		bool RestoreState();
		void ForgetState();

	private:
		CWnd* m_pDialog;
		CTypedPtrList< CPtrList, CControlMemory* >m_ControlList;
};
