// SnoopDat.cpp : implementation of the CSnoopView class
// See SnoopVw.cpp for other member functions of this class.
//

#include "stdafx.h"
#include "Snoop.h"

#include "SnoopDoc.h"
#include "SnoopVw.h"

#ifdef _DEBUG

#define ASSERT_RETURN(x)            if(!(x))            \
                                    {                   \
                                        ASSERT(FALSE);  \
                                        return;         \
                                    }

#define ASSERT_RETURN_VALUE(x,v)    if(!(x))            \
                                    {                   \
                                        ASSERT(FALSE);  \
                                        return v;       \
                                    }
#else

#define ASSERT_RETURN(x)
#define ASSERT_RETURN_VALUE(x,v)

#endif

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

class CSnoopString : public CString
{
public:
    // public function to access the
    //protected m_pchData of CString.
    LPBYTE GetSnoopData() { return (LPBYTE)m_pchData; }
};

void CSnoopView::FillClassCombo()
{
    m_comboClass.ResetContent();

    m_comboClass.AddString("CString");
    m_comboClass.AddString("CWordArray");
    m_comboClass.AddString("CStringArray");
    m_comboClass.AddString("CMapStringToString");
    // Add more classes/structures to the listbox
}

// Fills the Window bombobox with all the toplevel windows.
void CSnoopView::FillWindowCombo()
{
    m_comboWindow.ResetContent();

    // The mother of all windows.
    CWnd* pWnd = GetDesktopWindow();
    if (!pWnd)
    {
        ASSERT(FALSE);
        return;
    } 

    CString sText;
    CString sStr;

    pWnd = pWnd->GetWindow(GW_CHILD);
    while(pWnd)
    {
        if (pWnd->GetWindowTextLength() > 0)
        {
            pWnd->GetWindowText(sText);

            // Save the window handle to the right of the
            // text. Since the combobox cannot handle tab
            // characters (yet), add spaces before adding
            // the tab.
            sStr.Format("%s%150s\t%08X", sText, " ",
                                    pWnd->GetSafeHwnd());

            m_comboWindow.AddString(sStr);
        }
        pWnd = pWnd->GetWindow(GW_HWNDNEXT);
    }

    if (!m_sStaticWindowText.IsEmpty())
    {
        // Select the previously selected window - assuming
        // the programmer will be debugging the same program.
        int nIndex = m_comboWindow.FindString(
                                -1, m_sStaticWindowText);
        if (nIndex != -1)
        {
            m_comboWindow.SetCurSel(nIndex);
        }
    }
}

// Snoop button click handler.
void CSnoopView::OnBSnoop() 
{
    HANDLE hProcess = 0;
    LPBYTE pbyBase = 0;
    CString sClass;

    m_listData.ResetContent();

    // Get the input.
    if (!GetInput(pbyBase, sClass))
    {
        return;
    }

    if (sClass == "CString")
    {
        LPVOID pS = (LPVOID)malloc(sizeof CString);
        DisplayCString(pbyBase, pS);
        free(pS);
    }
    else
    if (sClass == "CWordArray")
    {
        LPVOID pWA = (LPVOID)malloc(sizeof CWordArray);
        DisplayCWordArray(pbyBase, pWA);
        free(pWA);
    }
    else
    if (sClass == "CStringArray")
    {
        LPVOID pSA = (LPVOID)malloc(sizeof CStringArray);
        DisplayCStringArray(pbyBase, pSA);
        free(pSA);
    }
    else
    if (sClass == "CMapStringToString")
    {
        LPVOID pMSS = (LPVOID)
                        malloc(sizeof CMapStringToString);
        DisplayCMapStringToString(pbyBase, pMSS);
        free(pMSS);
    }
    //
    // Add implementation of other classes here.
    //

    // Close the process.
    CloseHandle(m_hProcess);
}

// Gets the address, class and window.
BOOL CSnoopView::GetInput(LPBYTE& pBase, CString& sClass)
{
    // Get the address.
    CString sAddress;
    m_editAddress.GetWindowText(sAddress);
    ASSERT_RETURN_VALUE(m_editAddress.GetWindowTextLength()
                                                > 1,FALSE);

    // The address may have a "0x" prefix.
    int nStart = 0;
    if (sAddress.Left(2)=="0x" || sAddress.Left(2)=="0X")
        nStart = 2;

    pBase = 0;
    sscanf(sAddress.Mid(nStart), "%p", &pBase);
    if (!pBase)
    {
        AfxMessageBox("Invalid Address");
        return FALSE;
    }

    // Get the Class to snoop.
    int nIndex = m_comboClass.GetCurSel();
    ASSERT_RETURN_VALUE(nIndex>=0, FALSE);

    m_comboClass.GetLBText(nIndex, sClass);
    ASSERT_RETURN_VALUE(!sClass.IsEmpty(), FALSE);

    // Get Window handle.
    nIndex = m_comboWindow.GetCurSel();
    ASSERT_RETURN_VALUE(nIndex>=0, FALSE);

    m_comboWindow.GetLBText(nIndex, m_sStaticWindowText);
    ASSERT_RETURN_VALUE(!m_sStaticWindowText.IsEmpty(),
                                                FALSE);

    int nTab = m_sStaticWindowText.Find('\t');
    ASSERT_RETURN_VALUE(nTab!=-1, FALSE);

    // Get the window handle from the right end of the text
    HWND hWnd = 0;
    CString sText = m_sStaticWindowText.Mid(nTab+1);
    sscanf (sText, "%08X", &hWnd);
    ASSERT_RETURN_VALUE(hWnd, FALSE);

    // Confirm that the window is still around.
    if (!::IsWindow(hWnd))
    {
        AfxMessageBox("Invalid Window");
        return FALSE;
    }

    // get the window's process ID.
    DWORD dwProcessId = 0;
    GetWindowThreadProcessId(hWnd, &dwProcessId);
    if (!dwProcessId)
    {
        AfxMessageBox("Invalid ProcessId");
        return FALSE;
    }

    // Open the process for reading.
    m_hProcess = OpenProcess(PROCESS_VM_READ, TRUE,
                                            dwProcessId);
    if (!m_hProcess)
    {
        AfxMessageBox("Invalid ProcessId");
        return FALSE;
    }

    return TRUE;
}

BOOL CSnoopView::DisplayCString(LPBYTE pbyBase, LPVOID pS)
{
    if (!ReadMemory(pbyBase, pS, sizeof CString ))
    {
        return FALSE;
    }

    pbyBase = ((CSnoopString*)pS)->GetSnoopData();
    CString sData = GetCString(pbyBase);

    m_listData.AddString(sData);
    return TRUE;
}

BOOL CSnoopView::DisplayCWordArray(LPBYTE pbyBase,
                                        LPVOID pWA)
{
    class CSnoopWordArray : public CWordArray
    {
    public:
        // public functions to access the protected 
        // members of CWordArray.
        LPBYTE GetDataPtr() { return (LPBYTE)m_pData; }
        int GetMaxSize()    { return m_nMaxSize; }
        int GetGrowBy()     { return m_nGrowBy; }
    };

    if (!ReadMemory(pbyBase, pWA, sizeof CWordArray ))
    {
        return FALSE;
    }

    int nSize       = ((CSnoopWordArray*)pWA)->GetSize();
    int nMaxSize    = ((CSnoopWordArray*)pWA)->GetMaxSize();
    int nGrowBy     = ((CSnoopWordArray*)pWA)->GetGrowBy();

    CString sData;
    sData.Format("m_nSize\t\t= %d", nSize);
    m_listData.AddString(sData);

    sData.Format("m_nMaxSize\t= %d", nMaxSize);
    m_listData.AddString(sData);

    sData.Format("m_nGrowBy\t= %d", nGrowBy);
    m_listData.AddString(sData);

    m_listData.AddString("");
    m_listData.AddString("Data...");

    if (!nSize)
        return TRUE;

    pbyBase = ((CSnoopWordArray*)pWA)->GetDataPtr();
    LPWORD pbyArr = (LPWORD)malloc(nSize * sizeof WORD);

    if (!ReadMemory(pbyBase, (LPVOID)pbyArr,
                                    nSize * sizeof WORD))
    {
        free(pbyArr);
        m_listData.AddString("*** READ ERROR ***");
        return FALSE;
    }

    for (int nIndex=0; nIndex<nSize; nIndex++)
    {
        sData.Format("%d)\t%u", nIndex, pbyArr[nIndex]);
        m_listData.AddString(sData);
    }

    free(pbyArr);
    return TRUE;
}

BOOL CSnoopView::DisplayCStringArray(LPBYTE pbyBase,
                                                LPVOID pSA)
{
    class CSnoopStringArray : public CStringArray
    {
    public:
        // public functions to access the protected
        // members of CStringArray.
        LPBYTE GetDataPtr() { return (LPBYTE)m_pData; }
        int GetMaxSize()    { return m_nMaxSize; }
        int GetGrowBy()     { return m_nGrowBy; }
    };

    if (!ReadMemory(pbyBase, pSA, sizeof CStringArray ))
    {
        return FALSE;
    }

    // Get the protected members.
    int nSize     = ((CSnoopStringArray*)pSA)->GetSize();
    int nMaxSize  = ((CSnoopStringArray*)pSA)->GetMaxSize();
    int nGrowBy   = ((CSnoopStringArray*)pSA)->GetGrowBy();

    CString sData;
    sData.Format("m_nSize\t\t= %d", nSize);
    m_listData.AddString(sData);

    sData.Format("m_nMaxSize\t= %d", nMaxSize);
    m_listData.AddString(sData);

    sData.Format("m_nGrowBy\t= %d", nGrowBy);
    m_listData.AddString(sData);

    m_listData.AddString("");
    m_listData.AddString("Data...");

    if (!nSize)
        return TRUE;

    pbyBase = ((CSnoopStringArray*)pSA)->GetDataPtr();
    LPBYTE* pbyArr = (LPBYTE*)malloc(nSize*4);

    // Read the array of pointers.
    if (!ReadMemory(pbyBase, (LPVOID)pbyArr, nSize*4))
    {
        free(pbyArr);
        m_listData.AddString("*** READ ERROR ***");
        return FALSE;
    }

    // Loop thru and get each member of the array.
    CString sString;
    for (int nIndex=0; nIndex<nSize; nIndex++)
    {
        pbyBase = pbyArr[nIndex];

        sString = GetCString(pbyBase);
        sData.Format("%d)\t%s", nIndex, sString);

        m_listData.AddString(sData);
    }

    free(pbyArr);
    return TRUE;
}

BOOL CSnoopView::DisplayCMapStringToString(LPBYTE pbyBase,
                                                LPVOID pMSS)
{
    // Equivalent of the internal structure CAssoc
    // defined in CMapStringToString.
    struct CSnoopAssoc
    {
        CSnoopAssoc* pNext;
        UINT nHashValue;  // needed for efficient iteration
        CString key;
        CString value;
    };

    class CSnoopMapStringToString : public CMapStringToString
    {
    public:
        // public functions to access the protected
        // members of CMapStringToString.
        LPBYTE GetHashTablePtr()
            { return (LPBYTE)m_pHashTable; }
        UINT GetHashTableSize()
            { return m_nHashTableSize; }
    };

    // Read the CMapStringToString
    if (!ReadMemory(pbyBase, pMSS,
                        sizeof CMapStringToString ))
    {
        m_listData.AddString("*** READ ERROR ***");
        return FALSE;
    }

    // Get the protected hash table members.
    pbyBase =
        ((CSnoopMapStringToString*)pMSS)->GetHashTablePtr();
    UINT uHashTableSize =
        ((CSnoopMapStringToString*)pMSS)->GetHashTableSize();

    // Read the hash table pointers.
    LPVOID pvHash = (LPVOID)malloc(uHashTableSize*4);
    if (!ReadMemory(pbyBase, pvHash, uHashTableSize*4))
    {
        free(pvHash);
        m_listData.AddString("*** READ ERROR ***");
        return FALSE;
    }

    int nCount = 0;

    // Loop thru the hash table and get individual data.
    LPVOID pvAssoc = (LPBYTE)malloc (sizeof CSnoopAssoc);
    for (UINT uHash=0; uHash < uHashTableSize; uHash++)
    {
        pbyBase = (LPBYTE)((CSnoopAssoc**)pvHash)[uHash];
        while (pbyBase)
        {
            // read the internal CAssoc structure
            // which holds the data.
            if (!ReadMemory(pbyBase, pvAssoc,
                                        sizeof CSnoopAssoc))
            {
                m_listData.AddString("*** READ ERROR ***");
                break;
            }

            // Get the key string.
            LPVOID pvKey =
                (LPVOID) &((CSnoopAssoc*)pvAssoc)->key;
            LPBYTE pbyKey =
                ((CSnoopString*)pvKey)->GetSnoopData();
            CString sKey = GetCString(pbyKey);

            // Get the value string.
            LPVOID pvValue =
                (LPVOID) &((CSnoopAssoc*)pvAssoc)->value;
            LPBYTE pbyValue =
                ((CSnoopString*)pvValue)->GetSnoopData();
            CString sValue = GetCString(pbyValue);

            m_listData.AddString(sKey + "\t" + sValue);
            nCount++;

            pbyBase = (LPBYTE)((CSnoopAssoc*)pvAssoc)->pNext;
        }
    }

    free(pvAssoc);
    free(pvHash);

    // Add the number of elements to the top of the listbox
    CString sData;
    m_listData.InsertString(0, sData);
    sData.Format ("Number of Elements\t%d", nCount);
    m_listData.InsertString(0, sData);

    return TRUE;
}

// Gets the string and its parameters.
CString CSnoopView::GetCString(LPBYTE pbyBase)
{
    CString sString("*** Read Error ***");
    CStringData stringData;

    // First read the string parameters stored in CStringData
    LPBYTE pbyStringDataBase = pbyBase - sizeof CStringData;

    if (!ReadMemory(pbyStringDataBase,
                (LPBYTE)&stringData, sizeof CStringData ))
    {
        return sString;
    }

    long lRefs          = stringData.nRefs;
    int nDataLength     = stringData.nDataLength;
    int nAllocLength    = stringData.nAllocLength;

    CString sIncreaseSize;
    if (nDataLength)
    {
        CString sStr;
        int nSize = stringData.nDataLength + 1;
        LPSTR pszStr = sStr.GetBuffer(nSize);

        // Now we know the length. Read the actual string.
        if (!ReadMemory(pbyBase, pszStr, nSize))
        {
            lstrcpy(pszStr, "*** Read Error ***");
        }
        sStr.ReleaseBuffer();

        // return sStr if not interested in string parameters
        sString.Format("%d, %d, %d, %s", lRefs,
                            nDataLength, nAllocLength, sStr);
    }
    return sString;
}

// Wrapper around ReadProcessMemory.
BOOL CSnoopView::ReadMemory(LPBYTE pbyBase,
                                    LPVOID pOut, int nSize)
{
    DWORD dwReadCount = 0;

    if (!ReadProcessMemory(m_hProcess, pbyBase, pOut,
                                    nSize, &dwReadCount))
    {
        return FALSE;
    }

    if (dwReadCount != (DWORD)nSize)
    {
        return FALSE;
    }
    return TRUE;
}

